#include <armadillo>

using namespace arma;

#include "common.hpp"
#include "diff.cpp"

struct Spectrum {
    static const unsigned n_params = 4;
    
    Spectrum(double fq, unsigned r = 1, unsigned n = n_points) : x(n), y(n), d(n_params),
        freq(fq) {
        
        /* open the file */
        static char buffer[256];
        snprintf(buffer, 256, "samples/%d.%u", (int)freq, r);
        FILE* f = fopen(buffer, "r");
        assert(f);
    
        /* ignore the first lines */
        for(unsigned l = 0; l < 17; ++l) // +2 from the header
            fgets(buffer, 256, f);
            
        for(unsigned p = 0; p < n; ++p)
            fscanf( f, "%lf %lf", &y(p), &x(p) );
    
        fclose(f);
    }
    
    void print() {
        for(int i = 0, e = x.n_rows; i < e; ++i) {
            printf("%.12lf %.12lf %.12lf\n", x(i), y(i), fsine( x(i), d ));
        }
    }
    
    double calc_rms() {
        double sum = 0;
    
        for(int i = 0, e = y.n_rows; i < e; ++i)
            sum += y(i) * y(i);
    
        return sqrt(sum / y.n_rows);
    }
    
    double calc_peak() {
        double max = -1;
        
        for(int i = 0, e = y.n_rows; i < e; ++i) {
            double aux = fabs( y(i) );
            if( aux > max ) max = aux;
        }
        
        return max;
    }
    
    double calc_mse() {
        double acc = 0;
        
        for(int i = 0, e = x.n_rows; i < e; ++i) {
            double out = fsine( x(i), d );
            double err = y(i) - out;
            acc += err * err;
        }
        
        return acc / x.n_rows;
    }
    
    /* fits the data using the Levenberg-Marquardt algorithm */
    double fit() {
        d(0) = 0.7;   // amplitude
        d(1) = 0.1;   // phase
        d(2) = 0.01;  // offset
        d(3) = freq;  // freq
        
        double u = 0.1, v = 10;
        unsigned n = x.n_rows;
        mat J(n, n_params), E(n, 1), I = eye(n_params, n_params);
        
        for(unsigned it = 1; it <= 1e2; ++it) {
            for(unsigned i = 0; i < n; ++i) {
                J(i, 0) = fsine_diff_ampl  ( x(i), d );
                J(i, 1) = fsine_diff_phase ( x(i), d );
                J(i, 2) = 1;
                J(i, 3) = fsine_diff_freq  ( x(i), d );
                E(i, 0) = y.at(i) - fsine( x(i), d );
            }
            
            mat H        = J.t() * J;
            mat G        = J.t() * E;
            mat err_m    = E.t() * E;
            
            double l_err = err_m(0,0) / n;
            mat dlast    = d;
            d           += solve(H + I*u, G);
            double n_err = calc_mse();
            
            if(n_err < l_err) {
                u /= v;
                if( fabs(l_err - n_err) < 1e-6 ) break;
            } else {
                u *= v;
                d  = dlast;
            }
        }
        
        if(d(0) < 0) {
            d(0)  = -d(0);
            d(1) += M_PI;
        }
        
        return calc_mse();
    }
    
    vec x, y, d;
    double freq;
};

double work(double freq, int r) {
    Spectrum sp(freq, r);
    double err = sp.fit();
    
    double a1  = sp.calc_peak();
    double a2  = sp.calc_rms() * sqrt(2);
    double a3  = sp.d(0);
    
    printf("%lf %lf %lf %lf %lf %lf\n", freq, sp.d(3), a1, a2, a3, sp.d(2));
    
    return err;
}

int main(int argc, char* argv[]) {
    assert(argc == 2);
    double acc_err = 0;
    
    for(int i = 0; i < n_freqs; ++i)
        acc_err += work( freqs[i], atoi(argv[1]) );
        
    //fprintf(stderr, "error: %lf\n", sqrt(acc_err / n_freqs) );
    
    return 0;
}
